from contextlib import ContextDecorator
from django.core.exceptions import FieldError
from django.db.models.sql import query
from dataclasses import dataclass
from typing import Any
import functools

from garnett.fields import TranslatedField
from garnett.utils import get_current_language


# Save current join info
_JoinInfo = query.JoinInfo


@dataclass
class JoinInfo:
    final_field: Any
    targets: Any
    opts: Any
    joins: Any
    path: Any
    transform_function_func: Any

    @property
    def transform_function(self):
        if isinstance(self.final_field, TranslatedField):
            # If its a partial, it must have had a transformer applied - leave it alone!
            if isinstance(self.transform_function_func, functools.partial):

                return self.transform_function_func

            name = get_current_language()

            # Cloned in from django
            def transform(field, alias, *, name, previous):
                try:
                    wrapped = previous(field, alias)
                    return self.try_transform(wrapped, name)
                except FieldError:
                    # TODO: figure out how to handle this case as we don't have
                    # final_field or last_field_exception

                    # FieldError is raised if the transform doesn't exist.
                    # if isinstance(final_field, Field) and last_field_exception:
                    #     raise last_field_exception
                    # else:
                    #     raise
                    raise

            # -------------------

            return functools.partial(
                transform, name=name, previous=self.transform_function_func
            )

        return self.transform_function_func

    def try_transform(self, lhs, name):
        # Cloned in from django
        import difflib

        """
        Helper method for build_lookup(). Try to fetch and initialize
        a transform for name parameter from lhs.
        """
        transform_class = lhs.get_transform(name)
        if transform_class:
            return transform_class(lhs)
        else:
            output_field = lhs.output_field.__class__
            suggested_lookups = difflib.get_close_matches(
                name, output_field.get_lookups()
            )
            if suggested_lookups:
                suggestion = ", perhaps you meant %s?" % " or ".join(suggested_lookups)
            else:
                suggestion = "."
            raise FieldError(
                "Unsupported lookup '%s' for %s or join on the field not "
                "permitted%s" % (name, output_field.__name__, suggestion)
            )

    def __iter__(self):
        # Necessary to mimic a tuple
        for x in [
            self.final_field,
            self.targets,
            self.opts,
            self.joins,
            self.path,
            self.transform_function,
        ]:
            yield x


def apply_patches():
    """Apply monkey patches to django"""
    # This is needed to allow values/values_list and F lookups to work
    # The most dangerous monkey patch of all time
    query.JoinInfo = JoinInfo


def revert_patches():
    """Revert monkey patches to django"""
    query.JoinInfo = _JoinInfo


class patch_lookups(ContextDecorator):
    def __enter__(self):
        apply_patches()

    def __exit__(self, exc_type, exc_value, traceback):
        revert_patches()



from django.apps.registry import Apps
from django.db.migrations import RunPython
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from garnett.utils import get_current_language_code
import json
from typing import Callable, Dict, List


def _get_migrate_function(
    app_label: str,
    model_fields: Dict[str, List[str]],
    update: Callable[[str, str], str],
) -> Callable[[Apps, BaseDatabaseSchemaEditor], None]:
    """Generate a migration function given an update function for each value

    update is a function taking current language and old value and returning a new value
    """

    def migrate(apps: Apps, schema_editor: BaseDatabaseSchemaEditor) -> None:
        current_lang = get_current_language_code()

        for model_name, fields in model_fields.items():
            updated = []
            model = apps.get_model(app_label, model_name)
            for item in model.objects.all():
                for field_name in fields:
                    # Set new value retrieved from update function
                    value = getattr(item, field_name)
                    setattr(item, field_name, update(current_lang, value))
                updated.append(item)

            # Bulk update only the required fields
            model.objects.bulk_update(updated, fields)

    return migrate


def get_migration(app_label: str, model_fields: Dict[str, List[str]]) -> RunPython:
    """Generate a migration operation to prepare text fields for being translated

    Args:
        app_label: label for django app the migration is in
        model_fields: mapping of model name to list of text fields to migrate

    Returns:
        RunPython migration operation
    """

    def update_forwards(current_lang: str, value: str) -> str:
        return json.dumps({current_lang: value})

    def update_backwards(current_lang: str, value: str) -> str:
        return json.loads(value)[current_lang]

    return RunPython(
        code=_get_migrate_function(app_label, model_fields, update_forwards),
        reverse_code=_get_migrate_function(app_label, model_fields, update_backwards),
    )



from django.apps import AppConfig


class AppConfig(AppConfig):
    name = "garnett"

    def ready(self):
        from django.conf import settings

        if getattr(settings, "GARNETT_PATCH_REVERSION_COMPARE", False):
            from garnett.ext.reversion import patch_compare

            patch_compare()

        if getattr(settings, "GARNETT_PATCH_DJANGO_FILTERS", False):
            from garnett.ext.filters import patch_filters

            patch_filters()



from django.http import Http404
from django.utils.translation import gettext as _

import logging

from .utils import get_language_from_request, is_valid_language
from .context import set_field_language

logger = logging.getLogger(__name__)


class TranslationContextMiddleware:
    """
    This middleware catches the requested "garnett language" and:
     * sets a garnett language attribute on the request
     * defines a context variable that is used when reading or altering fields
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def validate(self, language):
        """Validate the language raising http errors if invalid"""
        return None

    def __call__(self, request):
        request.garnett_language = get_language_from_request(request)
        self.validate(request.garnett_language)
        with set_field_language(request.garnett_language):
            response = self.get_response(request)
            return response


class TranslationContextNotFoundMiddleware(TranslationContextMiddleware):
    """
    This middleware catches the requested "garnett language" and:
     * sets a garnett language attribute on the request
     * defines a context variable that is used when reading or altering fields
     * will raise a 404 if the request language is not in languages list
    """

    def validate(self, language):
        if not is_valid_language(language):
            lang_obj = language
            lang_name = lang_obj.display_name(language)
            lang_en_name = lang_obj.display_name()
            raise Http404(
                _("This server does not support %(lang_name)s" " [%(lang_en_name)s].")
                % {
                    "lang_name": lang_name,
                    "lang_en_name": lang_en_name,
                }
            )


class TranslationCacheMiddleware:
    """Middleware to cache the garnett language in the users session storage

    This must be after one of the above middlewares and after the session middleware
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if hasattr(request, "garnett_language") and hasattr(request, "session"):
            request.session["GARNETT_LANGUAGE_CODE"] = request.garnett_language
        else:
            logger.error(
                "TranslationCacheMiddleware must come after main garnett middleware "
                "and the session middleware."
            )

        return self.get_response(request)



from contextlib import ContextDecorator
import contextvars
from langcodes import Language

# Internal context var should be set via set_field_language and get via get_current_language
_ctx_language = contextvars.ContextVar("garnett_language")


class set_field_language(ContextDecorator):
    def __init__(self, language):
        if isinstance(language, Language):
            self.language = language
        else:
            self.language = Language.get(language)
        self.token = None

    def __enter__(self):
        self.token = _ctx_language.set(self.language)

    def __exit__(self, exc_type, exc_value, traceback):
        _ctx_language.reset(self.token)



# Patches for django-filters
"""
This has been added as django-filters is used by multiple libraries,
including DRF and Djgnao-Graphene.
"""

from django_filters.filterset import CharFilter
from garnett.fields import TranslatedField


class TranslatedCharFilter(CharFilter):
    pass


def patch_filters():
    from django_filters.filterset import FILTER_FOR_DBFIELD_DEFAULTS

    FILTER_FOR_DBFIELD_DEFAULTS.update(
        {TranslatedField: {"filter_class": TranslatedCharFilter}}
    )

    from django_filters.rest_framework.filterset import FILTER_FOR_DBFIELD_DEFAULTS

    FILTER_FOR_DBFIELD_DEFAULTS.update(
        {TranslatedField: {"filter_class": TranslatedCharFilter}}
    )



from django import template
from garnett.utils import lang_param
from langcodes import Language

register = template.Library()


@register.simple_tag(takes_context=True)
def switch_page_language(context, language_code):
    # http://stackoverflow.com/questions/2047622/how-to-paginate-django-with-other-get-variables
    request = context["request"]
    dict_ = request.GET.copy()
    dict_[lang_param()] = language_code
    get_params = dict_.urlencode()
    return f"{request.path}?{get_params}"


@register.filter
def language_display(language, display_language=None):
    if type(language) is str:
        language = Language.get(language)
    if display_language is None:
        return language.display_name()
    return language.display_name(display_language)


@register.inclusion_tag("garnett/language_selector.html", takes_context=True)
def language_selector(context, selector):
    context.update({"selector": selector})
    return {
        "selector": selector,
        "request": context["request"],
        "garnett_languages": context["garnett_languages"],
        "garnett_current_language": context["garnett_current_language"],
    }



from django.utils.translation import get_language

from garnett.utils import lang_param


def query(request):
    return request.GET.get(lang_param(), None)


def cookie(request):
    return request.COOKIES.get("GARNETT_LANGUAGE_CODE", None)


def session(request):
    return request.session.get("GARNETT_LANGUAGE_CODE", None)


def header(request):
    return request.META.get("HTTP_X_GARNETT_LANGUAGE_CODE", None)


def browser(request):
    return get_language()



from django.db.models import lookups
from django.db.models.fields import json, CharField
from django.db.models.functions import Cast
from django.contrib.postgres.fields.jsonb import KeyTextTransform
from django.contrib.postgres.lookups import SearchLookup, TrigramSimilar
from django.contrib.postgres.search import TrigramSimilarity

from garnett.fields import TranslatedField, TranslatedKeyTransform
from garnett.utils import get_current_language


# We duplicate and process_lhs and process_rhs to be certain these are
# Actually called during tests.
# Otherwise, the blank classes appear to give 100% coverage


@TranslatedField.register_lookup
class HasLang(json.HasKey):
    lookup_name = "has_lang"

    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class HasLangs(json.HasKeys):
    lookup_name = "has_langs"

    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class HasAnyLangs(json.HasAnyKeys):
    lookup_name = "has_any_langs"

    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


# Override default lookups on our field to handle language lookups


class CurrentLanguageMixin:
    """Mixin to perform language lookup on lhs"""

    def __init__(self, lhs, *args, **kwargs):
        tlhs = json.KeyTransform(
            str(get_current_language()),
            lhs,
        )
        super().__init__(tlhs, *args, **kwargs)


@TranslatedField.register_lookup
class BaseLanguageExact(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, lookups.Exact
):
    # Note: On some database engines lookup_name actually has an effect on the result
    # (See lookup_cast in the django postgres backend)
    lookup_name = "exact"
    prepare_rhs = False

    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageIExact(CurrentLanguageMixin, json.KeyTransformIExact):
    lookup_name = "iexact"

    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageIContains(CurrentLanguageMixin, json.KeyTransformIContains):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageContains(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, lookups.Contains
):
    # Override the default json field contains which is not a text contains
    # https://docs.djangoproject.com/en/3.1/topics/db/queries/#contains
    lookup_name = "contains"

    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


# Override contains lookup for after a key lookup i.e. title__en__contains="thing"
@TranslatedKeyTransform.register_lookup
class KeyTransformContains(json.KeyTransformTextLookupMixin, lookups.Contains):
    lookup_name = "contains"

    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedKeyTransform.register_lookup
class KeyTransformExact(json.KeyTransformTextLookupMixin, lookups.Exact):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedKeyTransform.register_lookup
class KeyTransformGreaterThan(json.KeyTransformTextLookupMixin, lookups.GreaterThan):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedKeyTransform.register_lookup
class KeyTransformGreaterThanOrEqual(
    json.KeyTransformTextLookupMixin, lookups.GreaterThanOrEqual
):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedKeyTransform.register_lookup
class KeyTransformLessThan(json.KeyTransformTextLookupMixin, lookups.LessThan):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedKeyTransform.register_lookup
class KeyTransformLesshanOrEqual(
    json.KeyTransformTextLookupMixin, lookups.LessThanOrEqual
):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageStartsWith(CurrentLanguageMixin, json.KeyTransformStartsWith):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageIStartsWith(CurrentLanguageMixin, json.KeyTransformIStartsWith):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageEndsWith(CurrentLanguageMixin, json.KeyTransformEndsWith):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageIEndsWith(CurrentLanguageMixin, json.KeyTransformIEndsWith):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageRegex(CurrentLanguageMixin, json.KeyTransformRegex):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageIRegex(CurrentLanguageMixin, json.KeyTransformIRegex):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageGreaterThan(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, lookups.GreaterThan
):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageGreaterThanOrEqual(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, lookups.GreaterThanOrEqual
):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageLessThan(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, lookups.LessThan
):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageLessThanOrEqual(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, lookups.LessThanOrEqual
):
    def process_lhs(self, compiler, connection):
        self.lhs = Cast(self.lhs, CharField())
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


@TranslatedField.register_lookup
class BaseLanguageSearch(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, SearchLookup
):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


# --- Postgres only functions ---


@TranslatedField.register_lookup
class BaseLanguageTrigramSimilar(
    CurrentLanguageMixin, json.KeyTransformTextLookupMixin, TrigramSimilar
):
    def process_lhs(self, compiler, connection):
        return super().process_lhs(compiler, connection)

    def process_rhs(self, compiler, connection):
        return super().process_rhs(compiler, connection)


class LangTrigramSimilarity(TrigramSimilarity):
    # There is no way we can calculate if a field is a language or not
    # So we have to write our own function here
    # We also need to cast the JSONB field to a string
    def __init__(self, expression, string, **extra):
        lang = str(get_current_language())
        expression = KeyTextTransform(lang, expression)
        super().__init__(expression, string, **extra)



from typing import List, Union

from django.conf import settings
from django.utils.module_loading import import_string
from django.core.exceptions import ImproperlyConfigured
from langcodes import Language

from garnett.context import _ctx_language


def lang_param():
    return getattr(settings, "GARNETT_QUERY_PARAMETER_NAME", "glang")


def get_default_language():
    setting = getattr(settings, "GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE", "en-AU")
    if callable(setting):
        default = setting()
    else:
        default = setting

    if isinstance(default, Language):
        return default
    elif isinstance(default, str):
        return Language.get(default)
    else:
        raise ImproperlyConfigured(
            "GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE must be a string or callable that returns a string or `Language` object"
        )


def codes_to_langs(content: dict) -> dict:
    return {Language.get(lang): text for lang, text in content.items()}


def get_property_name() -> str:
    return getattr(settings, "GARNETT_TRANSLATIONS_PROPERTY_NAME", "translations")


def normalise_language_code(langcode):
    return Language.get(langcode).to_tag()


def normalise_language_codes(value):
    return {normalise_language_code(lang): val for lang, val in value.items()}


def is_valid_language(language: Union[str, Language]) -> bool:
    if isinstance(language, Language):
        language = language
    if isinstance(language, str):
        language = Language.get(language)
    return language in get_languages()


def get_current_language() -> Language:
    lang = _ctx_language.get(None)
    if not lang:
        return get_default_language()
    return lang


def get_current_language_code() -> str:
    return get_current_language().to_tag()


def get_languages() -> List[Language]:
    langs = getattr(
        settings, "GARNETT_TRANSLATABLE_LANGUAGES", [get_default_language()]
    )
    if callable(langs):
        langs = langs()
    if type(langs) == list:
        return [Language.get(lang) for lang in langs]
    raise ImproperlyConfigured(
        "GARNETT_TRANSLATABLE_LANGUAGES must be a list or a callable that returns a list"
    )


def get_language_from_request(request) -> Language:
    opt_order = getattr(
        settings,
        "GARNETT_REQUEST_LANGUAGE_SELECTORS",
        [
            "garnett.selectors.header",
            "garnett.selectors.query",
            "garnett.selectors.cookie",
        ],
    )
    for opt in opt_order:
        func = import_string(opt)
        if lang := func(request):
            return Language.get(lang)
    return get_default_language()



from .utils import get_languages, get_current_language


def languages(request):
    return {
        "garnett_languages": get_languages(),
        "garnett_current_language": get_current_language(),
    }



class LanguageStructureError(Exception):
    """
    A language structure isn't a dictionary
    """

    pass


class LanguageContentError(Exception):
    """
    A language in a language dictionary has non-string content
    """

    pass


class LanguageSelectionError(Exception):
    """
    A language in a language dictionary has non-string content
    """

    pass



from garnett.fields import TranslatedField


def patch_compare():
    from reversion_compare.compare import CompareObjects as COBase

    class CompareObjects(COBase):
        def __init__(self, field, field_name, obj, version1, version2, is_reversed):
            if isinstance(field, TranslatedField):
                field_name = field.attname
            super().__init__(field, field_name, obj, version1, version2, is_reversed)

    # Import this as late as possible as we are monkey patching over it
    import reversion_compare.compare

    reversion_compare.compare.CompareObjects = CompareObjects



from django.core.serializers import json
from garnett.serializers.base import TranslatableSerializer


class Serializer(TranslatableSerializer, json.Serializer):
    pass


def Deserializer(stream_or_string, **options):
    yield from json.Deserializer(stream_or_string, **options)



from typing import Tuple, Optional, Callable
from django.utils.translation import gettext as _

from langcodes import Language
from garnett.utils import (
    codes_to_langs,
    get_current_language,
    get_current_language_code,
    get_languages,
)
from garnett import exceptions as e


class HTMLTranslationMixin:
    def __html__(self) -> str:
        # Add leading [lang] wrapped in a span
        text = self
        if not self.is_fallback:
            return text
        elif not self.fallback_language:
            return "??"
        else:
            return (
                '<span class="fallback"'
                'data-lang-code="{lang}"'
                ">"
                "[{lang}]</span> "
                "{s}"
            ).format(s=self, lang=self.fallback_language.to_tag())


class TranslatedStr(str):
    """
    A translated string subclasses string and allows us to attach more information about
    how a string was generated and the language of the string.
    """

    def __new__(cls, content, fallback: Callable = None):
        try:
            current_language_code = get_current_language_code()
            has_current_language = current_language_code in content.keys()
        except (AttributeError, TypeError):
            raise e.LanguageStructureError

        if has_current_language:
            fallback_language = None
            text = content.get(current_language_code)
        else:
            if fallback:
                fallback_language, text = fallback(content)
            else:
                fallback_language, text = cls.get_fallback_text(content)

        instance = super().__new__(cls, text)
        instance.translations = codes_to_langs(content)
        instance.is_fallback = not has_current_language
        instance.fallback_language = fallback_language
        return instance

    @classmethod
    def get_fallback_text(cls, content) -> Tuple[Optional[Language], str]:
        return None, ""


class VerboseTranslatedStr(TranslatedStr):
    """
    A translated string that gives information if a string isn't present.
    """

    @classmethod
    def get_fallback_text(cls, content):
        """Default fallback function that returns an error message"""
        language = get_current_language()
        lang_name = language.display_name(language)
        lang_en_name = language.display_name()

        return language, content.get(
            language.to_tag(),
            _(
                "No translation of this field available in %(lang_name)s"
                " [%(lang_en_name)s]."
            )
            % {
                "lang_name": lang_name,
                "lang_en_name": lang_en_name,
            },
        )


class NextTranslatedStr(TranslatedStr, HTMLTranslationMixin):
    """
    A translated string that fallsback based on the order of preferred langages in the app.
    """

    @classmethod
    def get_fallback_text(cls, content):
        """Fallback that checks each language consecutively"""
        for lang in get_languages():
            if lang.to_tag() in content:
                return lang, content[lang.to_tag()]

        return None, ""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['garnett', 'garnett.ext', 'garnett.serializers', 'garnett.templatetags']

package_data = \
{'': ['*'], 'garnett': ['templates/admin/*', 'templates/garnett/*']}

install_requires = \
['django>=3.1.0,<3.2.0', 'langcodes>=2.1,<2.2', 'language-data>=1.0.1,<1.1.0']

setup_kwargs = {
    'name': 'django-garnett',
    'version': '0.4.3',
    'description': 'Simple translatable Django fields',
    'long_description': '# django-garnett\n\nDjango Garnett is a field level translation library that allows you to store strings in multiple languages for fields in Django - with minimal changes to your models and without having to rewrite your code.\n\nWant a demo? https://django-garnett.herokuapp.com/\n\n<a href="https://www.aristotlemetadata.com"\n    style="text-decoration: none">\nMade with <i style="color:#d63384">♥</i> by\n<img src="https://brand.aristotlemetadata.com/images/svgs/thick.svg" height=20> Aristotle Metadata\n</a>\n\nIn summary it allows you to do this:\n\n<table>\n<tr>\n    <td>\n        <tt>models.py</tt>\n    <td>\n        You can do this!\n</tr>\n<tr>\n<td>\nBy changing your models from this...\n\n```python\nclass Greeting(models.model):\n    text = CharField(max_length=150)\n    target = models.CharField()\n    def __str__(self):\n        return f"{self.greeting}, {self.target}"\n\n```\n\nto this...\n\n```python\n# Import garnett\nfrom garnett.fields import Translated\n\nclass Greeting(models.model):\n    # Convert greeting to a translatable field\n    text = Translated(CharField(max_length=150))\n    target = models.CharField()\n    def __str__(self):\n        return f"{self.greeting} {self.target}"\n```\n</td>\n<td>\n\n```python\nfrom garnett.context import set_field_language\ngreeting = Greeting(text="Hello", target="World")\n\nwith set_field_language("en"):\n    greeting.text = "Hello"\nwith set_field_language("fr"):\n    greeting.text = "Bonjour"\n\ngreeting.save()\ngreeting.refresh_from_db()\n\nwith set_field_language("en"):\n    print(greeting.text)\n    print(greeting)\n# >>> "Hello"\n# >>> "Hello World"\n\nwith set_field_language("fr"):\n    print(greeting.text)\n    print(greeting)\n# >>> "Bonjour"\n# >>> "Bonjour World!"\n\nwith set_field_language("en"):\n    print(greeting.text)\n    print(greeting)\n# >>> "Hello"\n# >>> "Hello World"\n    Greeting.objects.filter(title="Hello").exists()\n# >>> True\n    Greeting.objects.filter(title="Bonjour").exists()\n# >>> False\n    Greeting.objects.filter(title__fr="Bonjour").exists()\n# >>> True!!\n\n# Assuming that GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE="en"\n# Or a middleware has set the language context\nprint(greeting.text)\n# >>> Hello\nprint(greeting)\n# >>> Hello World!\n\n```\n\n</td>\n</table>\n\nTested on:\n  - Django 3.1+\n  - Postgres, SQLite, MariaDB\n  - Python 3.7+\n\nTested with the following libraries:\n  - [Django Rest Framework](https://www.django-rest-framework.org/)\n  - [django-reversion](https://github.com/etianen/django-reversion) & [django-reversion-compare](https://github.com/jedie/django-reversion-compare)\n  - [django-filters](https://github.com/carltongibson/django-filter)\n\nPros:\n* Battletested in production - [Aristotle Metadata](https://www.aristotlemetadata.com) built, support and uses this library for 2 separate products, served to government and enterprise clients!\n* Fetching all translations for a model requires a single query\n* Translations are stored in a single database field with the model\n* All translations act like a regular field, so `Model.field_name = "some string"` and `print(Model.field_name)` work as you would expect\n* Includes a configurable middleware that can set the current language context based on users cookies, query string or HTTP headers\n* Works nicely with Django Rest Framework - translatable fields can be set as strings or as json dictionaries\n* Works nicely with Django `F()` and `Q()` objects within the ORM - and when it doesn\'t we have a language aware `LangF()` replacement you can use.\n\nCons:\n* You need to alter the models, so you can\'t make third-party libraries translatable.\n* It doesn\'t work on `queryset.values_list` - but we have a workaround below.\n\n## Why write a new Django field translator?\n\nA few reasons:\n* Most existing django field translation libraries are static, and add separate database columns or extra tables per translation.\n* Other libraries may not be compatible with common django libraries, like django-rest-framework.\n* We had a huge codebase that we wanted to upgrade to be multilingual - so we needed a library that could be added in without requiring a rewriting every access to fields on models, and only required minor tweaks.\n\nNote: Field language is different to the django display language. Django can be set up to translate your pages based on the users browser and serve them with a user interface in their preferred language.\n\nGarnett *does not* use the browser language by design - a user with a French browser may want the user interface in French, but want to see content in English or French based on their needs.\n\n\n# How to install\n\n1. Add `django-garnett` to your dependencies. eg. `pip install django-garnett`\n2. Convert your chosen field using the `Translated` function\n\n    * For example: `title = fields.Translated(models.CharField(*args))`\n\n3. Add `GARNETT_TRANSLATABLE_LANGUAGES` (a callable or list of [language codes][term-language-code]) to your django settings.\n    > Note: At the moment there is no way to allow "a user to enter any language".\n4. Add `GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE` (a callable or single [language code][term-language-code]) to your settings.\n5. Re-run `django makemigrations` & `django migrate` for any apps you\'ve updated.\n6. Thats mostly it.\n\nYou can also add a few optional value adds:\n\n7. (Optional) Add a garnett middleware to take care of field language handling:\n\n    * You want to capture the garnett language in a context variable available in views use: `garnett.middleware.TranslationContextMiddleware`\n\n    * You want to capture the garnett language in a context variable available in views, and want to raise a 404 if the user requests an invalid language use: `garnett.middleware.TranslationContextNotFoundMiddleware`\n\n    * (Future addition) You want to capture the garnett language in a context variable available in views, and want to redirect to the default language if the user requests an invalid language use: `garnett.middleware.TranslationContextRedirectDefaultMiddleware`\n\n    * If you want to cache the current language in session storage use `garnett.middleware.TranslationCacheMiddleware` after one of the above middleware (this is useful with the session selector mentioned below)\n\n8. (Optional) Add the `garnett` app to your `INSTALLED_APPS` to use garnett\'s template_tags. If this is installed before `django.contrib.admin` it also include a language switcher in the Django Admin Site.\n\n9. (Optional) Add a template processor:\n\n    * Install `garnett.context_processors.languages` this will add `garnett_languages` (a list of available `Language`s) and `garnett_current_language` (the currently selected language).\n\n10. (Optional) Add a custom translation fallback:\n\n    By default, if a language isn\'t available for a field, Garnett will show a mesage like:\n    > No translation of this field available in English\n\n    You can override this either by creating a custom fallback method:\n    ```\n    Translated(CharField(max_length=150), fallback=my_fallback_method))\n    ```\n    Where `my_fallback_method` takes a dictionary of language codes and corresponding strings, and returns the necessary text.\n\n    Additionally, you can customise how django outputs text in templates by creating a new\n    `TranslationStr` class, and overriding the [`__html__` method][dunder-html].\n\n\n\n## `Language` vs language\n\nDjango Garnett uses the python `langcodes` library to determine more information about the languages being used - including the full name and local name of the language being used. This is stored as a `Language` object.\n\n\n## Django Settings options:\n\n* `GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE`\n    * Stores the default language to be used for reading and writing fields if no language is set in a context manager or by a request.\n    * By default it is \'en-AU\' the [language code][term-language-code] for \'Strayan, the native tongue of inhabitants of \'Straya (or more commonly known as Australia). \n    * This can also be a callable that returns list of language codes. Combined with storing user settings in something like (django-solo)[https://github.com/lazybird/django-solo] users can dynamically add or change their language settings.\n    * default: `\'en-AU\'`\n* `GARNETT_TRANSLATABLE_LANGUAGES`:\n    * Stores a list of [language codes][term-language-code] that users can use to save against TranslatableFields.\n    * This can also be a callable that returns list of language codes. Combined with storing user settings in something like (django-solo)[https://github.com/lazybird/django-solo] users can dynamically add or change their language settings.\n    * default `[GARNETT_DEFAULT_TRANSLATABLE_LANGUAGE]`\n* `GARNETT_REQUEST_LANGUAGE_SELECTORS`:\n    * A list of string modules that determines the order of options used to determine the language selected by the user. The first selector found is used for the language for the request, if none are found the DEFAULT_LANGUAGE is used. These can any of the following in any order:\n        * `garnett.selector.query`: Checks the `GARNETT_QUERY_PARAMETER_NAME` for a language to display\n        * `garnett.selector.cookie`: Checks for a cookie called `GARNETT_LANGUAGE_CODE` for a language to display.\n            Note: you cannot change this cookie name.\n        * `garnett.selector.session`: Checks for a session key `GARNETT_LANGUAGE_CODE` for a language to display.\n            Note: you cannot change this key name.\n        * `garnett.selector.header`: Checks for a HTTP Header called `X-Garnett-Language-Code` for a language to display.\n            Note: you cannot change this Header name.\n        * `garnett.selector.browser`: Uses Django\'s `get_language` function to get the users browser/UI language [as determined by Django][django-how].\n    * For example, if you only want to check headers and cookies in that order, set this to `[\'garnett.selectors.header\', \'garnett.selectors.cookie\']`.\n    * default: `[\'garnett.selectors.header\', \'garnett.selectors.query\', \'garnett.selectors.cookie\']`\n* `GARNETT_QUERY_PARAMETER_NAME`:\n    * The query parameter used to determine the language requested by a user during a HTTP request.\n    * default: `glang`\n\nAdvanced Settings (you probably don\'t need to adjust these)\n* `GARNETT_TRANSLATABLE_FIELDS_PROPERTY_NAME`:\n    * Garnett adds a property to all models that returns a list of all TranslatableFields. By default, this is \'translatable_fields\', but you can customise it here if you want.\n    * default: `translatable_fields`\n* `GARNETT_TRANSLATIONS_PROPERTY_NAME`:\n    * Garnett adds a property to all models that returns a dictionary of all translations of all TranslatableFields. By default, this is \'translations\', but you can customise it here if you want.\n    * default: `translations`\n\n# Using Garnett\n\nIf you did everything above correctly, garnett should for the most part "just work".\n\n## Switching the active language\n\nGarnett comes with a handy context manager that can be used to specify the current language. In any place where you want to manually control the current language, wrap your code in `set_field_language` and garnett will correctly store the language. This can be nested, or you can change the language for a context multiple times before saving.\n\n```python\nfrom garnett.context import set_field_language\ngreeting = Greeting(text="Hello", target="World")\n\nwith set_field_language("en"):\n    greeting.text = "Hello"\nwith set_field_language("fr"):\n    greeting.text = "Bonjour"\n\ngreeting.save()\n```\n\n## Using Garnett with `values_list`\n\nThis is one of the areas that garnett _doesn\'t_ work immediately, but there is a solution.\n\nIn the places you are using values lists, wrap any translated field in an L-expression and the values list will return correctly. For example:\n\n```python\nfrom garnett.expressions import L\nBook.objects.values_list(L("title"))\n```\n\n## Using Garnett with Django-Rest-Framework\n\nAs `TranslationField`s are based on JSONField, by default Django-Rest-Framework renders these as a JSONField, which may not be ideal.\n\nYou can get around this by using the `TranslatableSerializerMixin` _as the first mixin_, which adds the necessary hooks to your serializer. This will mean class changes, but you won\'t need to update or override every field.\n\nFor example:\n\n```\nfrom rest_framework import serializers\nfrom library_app import models\nfrom garnett.ext.drf import TranslatableSerializerMixin\n\n\nclass BookSerializer(TranslatableSerializerMixin, serializers.ModelSerializer):\n    class Meta:\n        model = models.Book\n        fields = "__all__"\n```\n\nThis will allow you to set the value for a translatable as either a string for the active langauge, or by setting a dictionary that has all languages to be saved (note: this will override the existing language set).\nFor example:\n\nTo override just the active language: \n\n    curl -X PATCH ... -d "{  \\"title\\": \\"Hello\\"}"\n\nTo specifically override a single language (for example, Klingon): \n\n    curl -X PATCH ...  -H  "X-Garnett-Language-Code: tlh" -d "{  \\"title\\": \\"Hello\\"}"\n\nTo override all languages:\n\n    curl -X PATCH ... -d "{  \\"title\\": {\\"en\\": \\"Hello\\", \\"fr\\": \\"Bonjour\\"}}"\n\n## Using Garnett with django-reversion and django-reversion-compare\n\nThere are a few minor tweaks required to get Garnett to operate properly with\ndjango-reversion and django-reversion-compare based on how they serialise and display data.\n\nThis is because Garnett does not use the same \'field.attname\' and \'field.name\' which means serialization in Django will not work correctly.\n\nTo get django-reversion to work you will need to use a translation-aware serialiser and apply a patch to ensure that django-reversion-compare can show the right information.\n\nAn example json translation-aware serializer is included with Garnett and this can be applied with the following two settings in `settings.py`:\n\n```\n# In settings.py\n\nGARNETT_PATCH_REVERSION_COMPARE = True\nSERIALIZATION_MODULES = {"json": "garnett.serializers.json"}\n```\n\nTranslatedFields will list the history and changes in json, but it does do comparisons correctly.\n\n## Why call it Garnett?\n\n* Libraries need a good name.\n* Searching for "Famous Translators" will tell you about [Constance Garnett](https://en.wikipedia.org/wiki/Constance_Garnett).\n* Searching for "Django Garnett" showed there was no python library with this name.\n* It did however talk about [Garnet Clark](https://en.wikipedia.org/wiki/Garnet_Clark) (also spelled Garnett), a jazz pianist who played with Django Reinhart - the namesake of the Django Web Framework.\n* Voila - a nice name\n\n## Warnings\n\n* `contains` acts like `icontains` on SQLite only when doing a contains query\n it does a case insensitive search. I don\'t know why - https://www.youtube.com/watch?v=PgGNWRtceag\n* Due to how django sets admin form fields you will not get the admin specific widgets like\n  `AdminTextAreaWidget` on translated fields in the django admin site by default. They can however\n  be specified explicitly on the corresponding admin model form.\n\n## Want to help maintain this library?\n\nThere is a `/dev/` directory with a docker-compose stack you can ues to bring up a database and clean development environment.\n\n## Want other options?\n\nThere are a few good options for adding translatable strings to Django that may meet other use cases. We\'ve included a few other options here, their strengths and why we didn\'t go with them.\n\n* [django-modeltranslation](https://github.com/deschler/django-modeltranslation)\n  \n   **Pros:** this library lets you apply translations to external apps, without altering their models.\n   \n   **Cons:** each translation adds an extra column, which means languages are specified in code, and can\'t be altered by users later.\n* [django-translated-fields](https://github.com/matthiask/django-translated-fields)\n  \n   **Pros:** uses a great context processor for switching lanugages (which is where we got the idea for ours).\n   \n   **Cons:** Languages are specified in django-settings, and can\'t be altered by users later.\n* [django-parler](https://github.com/django-parler/django-parler)\n  \n   **Pros:** Django admin site support.\n   \n   **Cons:** Languages are stored in a separate table and can\'t be altered by users later. Translated fields are specified in model meta, away from the fields definition which makes complex lookups harder.\n\n\n\n[term-language-code]: https://docs.djangoproject.com/en/3.1/topics/i18n/#term-language-code\n[django-how]: https://docs.djangoproject.com/en/3.1/topics/i18n/translation/#how-django-discovers-language-preference\n[dunder-html]: https://code.djangoproject.com/ticket/7261',
    'author': 'Aristotle Metadata Enterprises',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from django.core.serializers.base import Serializer as BaseSerializer
from django.utils.encoding import is_protected_type
from garnett.fields import TranslatedField


def fetch_item_and_convert_to_generator(queryset):
    # This method exists so we can get the first item of an iterable
    # (which can include a generator), and then still iterate through the entire
    # set later.
    # This ensures that if we have a queryset we only fetch it once,
    # or run through the generator once.
    inner_iterable = iter(queryset)

    # Get the first item so we can do introspection
    item = next(inner_iterable)

    # Make a generator that returns the first item, then the rest of the iterable
    def inner_generator():
        yield item
        yield from inner_iterable

    return item, inner_generator()


class TranslatableSerializer(BaseSerializer):
    def serialize(
        self,
        queryset,
        *,
        stream=None,
        fields=None,
        use_natural_foreign_keys=False,
        use_natural_primary_keys=False,
        progress_output=None,
        object_count=0,
        **options
    ):

        if fields is not None:
            item, queryset = fetch_item_and_convert_to_generator(queryset)
            selected_fields = []
            for f in item._meta.fields:
                if f.name in fields:
                    if isinstance(f, TranslatedField):
                        selected_fields.append(f.attname)
                    else:
                        selected_fields.append(f.name)
            fields = selected_fields

        return super().serialize(
            queryset,
            stream=stream,
            fields=fields,
            use_natural_foreign_keys=use_natural_foreign_keys,
            use_natural_primary_keys=use_natural_primary_keys,
            progress_output=progress_output,
            object_count=object_count,
            **options
        )

    def _value_from_field(self, obj, field):
        value = field.value_from_object(obj)
        # Protected types (i.e., primitives like None, numbers, dates,
        # and Decimals) are passed through as is. All other values are
        # converted to string first.
        if isinstance(field, TranslatedField):
            return field.translations_from_object(obj)
        return value if is_protected_type(value) else field.value_to_string(obj)



"""
Helper methods for working with django-rest-framework
"""

from enum import Enum
from rest_framework.fields import JSONField, empty
from rest_framework import serializers

from garnett.fields import TranslatedField


class LanguageTypes(Enum):
    monolingual = 1
    multilingual = 2


class TranslatableAPIField(JSONField):
    def __init__(self, *args, **kwargs):
        self.innerfield = kwargs.pop("innerfield", None)
        super().__init__(*args, **kwargs)

    @property
    def validators(self):
        if not hasattr(self, "_validators"):
            self._validators = self.get_validators()

        if self.translation_type == LanguageTypes.monolingual:
            return self.innerfield.validators
        else:
            return self._validators

    @validators.setter
    def validators(self, validators):
        self._validators = validators

    def run_validation(self, data=empty):
        (is_empty_value, data) = self.validate_empty_values(data)
        if is_empty_value:
            return data
        value = self.to_internal_value(data)

        if isinstance(data, str):
            self.translation_type = LanguageTypes.monolingual
        elif isinstance(data, dict):
            self.translation_type = LanguageTypes.multilingual
        else:
            # This branch shouldn't occur, but we'll let it pass.
            # This will get caught in the main validators.
            self.translation_type = LanguageTypes.multilingual
        self.run_validators(value)
        return value


translatable_serializer_field_mapping = {TranslatedField: TranslatableAPIField}
translatable_serializer_field_mapping.update(
    serializers.ModelSerializer.serializer_field_mapping.copy()
)


class TranslatableSerializerMixin:
    serializer_field_mapping = translatable_serializer_field_mapping

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super().build_standard_field(
            field_name, model_field
        )
        if field_class is TranslatableAPIField:
            field_kwargs["innerfield"] = model_field.field
        return field_class, field_kwargs



from django.conf import settings
from django.core import exceptions
from django.db.models import JSONField
from django.db.models.fields.json import KeyTransform
from dataclasses import make_dataclass
from functools import partial
import logging
from typing import Callable, Dict, Union

from garnett.translatedstr import TranslatedStr, VerboseTranslatedStr
from garnett.utils import (
    get_current_language_code,
    get_property_name,
    get_languages,
    is_valid_language,
    normalise_language_codes,
)

logger = logging.getLogger(__name__)


def innerfield_validator_factory(innerfield) -> callable:
    def validator(values: dict):
        from garnett import exceptions as e

        if not isinstance(values, dict):
            raise e.LanguageStructureError(
                "Invalid value assigned to translatable field"
            )

        # Run validators on sub field
        errors = []
        for code, value in values.items():
            # Check language codes
            if not isinstance(value, str):
                raise exceptions.ValidationError(f'Invalid value for language "{code}"')
            if not is_valid_language(code):
                raise exceptions.ValidationError(
                    f'"{code}" is not a valid language code'
                )

            for v in innerfield.validators:
                try:
                    v(value)
                except exceptions.ValidationError as e:
                    errors.extend(e.error_list)
        if errors:
            raise exceptions.ValidationError(errors)

    return validator


def translatable_default(
    inner_default: Union[str, Callable[[], str]]
) -> Dict[str, str]:
    """Return default from inner field as dict with current language"""
    lang = get_current_language_code()
    if callable(inner_default):
        return {lang: inner_default()}

    return {lang: inner_default}


class TranslatedField(JSONField):
    """Translated text field that mirrors the behaviour of another text field

    All arguments except fallback can be provided on the inner field
    """

    def __init__(self, field, *args, fallback=None, **kwargs):
        self.field = field
        self._fallback = fallback

        if type(fallback) is type and issubclass(fallback, TranslatedStr):
            self.fallback = fallback
        elif callable(fallback):
            self.fallback = partial(TranslatedStr, fallback=fallback)
        else:
            self.fallback = VerboseTranslatedStr

        # Move some args to outer field
        outer_args = [
            "db_column",
            "db_index",
            "db_tablespace",
            "help_text",
            "verbose_name",
        ]
        inner_kwargs = self.field.deconstruct()[3]
        for arg_name in outer_args:
            if arg_name in inner_kwargs:
                kwargs[arg_name] = inner_kwargs[arg_name]

        # Create default for outer field based on inner field
        if "default" not in kwargs and "default" in inner_kwargs:
            # Use partial because it is serializable in django migrations
            kwargs["default"] = partial(translatable_default, inner_kwargs["default"])

        super().__init__(*args, **kwargs)
        self.validators.append(innerfield_validator_factory(self.field))

    def formfield(self, **kwargs):
        # We need to bypass the JSONField implementation
        return self.field.formfield(**kwargs)

    def get_prep_value(self, value):
        if hasattr(value, "items"):
            value = {
                lang_code: self.field.get_prep_value(text)
                for lang_code, text in value.items()
            }
        elif type(value) == str:
            value = self.field.get_prep_value(value)
            return value
        return super().get_prep_value(value)

    def from_db_value(self, value, expression, connection):
        value = super().from_db_value(value, expression, connection)
        if hasattr(self.field, "from_db_value"):
            value = {
                k: self.field.from_db_value(v, expression, connection)
                for k, v in value.items()
            }
        return value

    def value_from_object(self, obj):
        """Return the value of this field in the given model instance."""
        all_ts = getattr(obj, self.ts_name)
        if type(all_ts) is not dict:
            logger.warning(
                "Displaying an untranslatable field - model:{} (pk:{}), field:{}".format(
                    type(obj), obj.pk, self.name
                )
            )
            return str(all_ts)

        language = get_current_language_code()
        return all_ts.get(language, None)

    def translations_from_object(self, obj):
        """Return the value of this field in the given model instance."""
        return getattr(obj, self.ts_name)

    def get_attname(self):
        # Use field with _tsall as the attribute name on the object
        # We've overrided this as this is usually the name of the attribute used to populate the database
        return self.name + "_tsall"

    def get_attname_column(self):
        attname = self.get_attname()
        # Use name without _tsall as the column name
        column = self.db_column or self.name
        return attname, column

    @property
    def ts_name(self):
        return f"{self.name}_tsall"

    def contribute_to_class(self, cls, name, private_only=False):
        super().contribute_to_class(cls, name, private_only)

        # We use `ego` to differentiate scope here as this is the inner self
        # Maybe its not necessary, but it is funny.

        @property
        def translator(ego):
            return self.fallback(getattr(ego, self.ts_name))

        @translator.setter
        def translator(ego, value):
            """Setter for main field (without _tsall)"""
            all_ts = getattr(ego, self.ts_name)
            if not all_ts:
                # This is probably the first save through
                all_ts = {}
            elif type(all_ts) is not dict:
                logger.warning(
                    "Saving a broken field - model:{} (pk:{}), field:{}".format(
                        type(ego), ego.pk, name
                    )
                )
                logger.debug("Field data was - {}".format(all_ts))
                all_ts = {}

            if isinstance(value, str):
                language_code = get_current_language_code()
                all_ts[language_code] = value
            elif value is None:
                language_code = get_current_language_code()
                all_ts[language_code] = ""
            elif isinstance(value, dict):
                # normalise all language codes
                all_ts = normalise_language_codes(value)
            else:
                bad_type = type(value)
                raise TypeError(
                    f"Invalid type assigned to translatable field - {bad_type}"
                )

            setattr(ego, self.ts_name, all_ts)

        setattr(cls, f"{name}", translator)

        # This can probably be cached on the class
        @property
        def translatable_fields(ego):
            return [
                field
                for field in ego._meta.get_fields()
                if isinstance(field, TranslatedField)
            ]

        propname = getattr(
            settings, "GARNETT_TRANSLATABLE_FIELDS_PROPERTY_NAME", "translatable_fields"
        )
        setattr(cls, propname, translatable_fields)

        @property
        def translations(ego):
            translatable_fields = ego.translatable_fields
            Translations = make_dataclass(
                "Translations", [(f.name, dict) for f in translatable_fields]
            )
            kwargs = {}
            for field in translatable_fields:
                kwargs[field.name] = getattr(ego, field.ts_name)
            return Translations(**kwargs)

        setattr(cls, get_property_name(), translations)

        @property
        def available_languages(ego):
            """Returns a list of codes available on the whole model"""
            langs = set()
            for field in ego.translatable_fields:
                langs |= getattr(ego, field.ts_name, {}).keys()
            return [lang for lang in get_languages() if lang.language in langs]

        setattr(cls, "available_languages", available_languages)

    def get_transform(self, name):
        # Call back to the Field get_transform
        transform = super(JSONField, self).get_transform(name)
        if transform:
            return transform
        # Use our new factory
        return TranslatedKeyTransformFactory(name)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        args.insert(0, self.field)
        kwargs["fallback"] = self._fallback
        return name, path, args, kwargs


class TranslatedKeyTransform(KeyTransform):
    """Key transform for translate fields

    so we can register lookups on this without affecting the regular json field
    """


class TranslatedKeyTransformFactory:
    def __init__(self, key_name):
        self.key_name = key_name

    def __call__(self, *args, **kwargs):
        return TranslatedKeyTransform(self.key_name, *args, **kwargs)


# Shorter name for the class
Translated = TranslatedField


# Import lookups here so that they are registered by just importing the field
from garnett import lookups  # noqa: F401, E402



from django.db.models import F
from django.db.models.fields.json import KeyTextTransform

from garnett.utils import get_current_language
from garnett.fields import TranslatedField


# Based on: https://code.djangoproject.com/ticket/29769#comment:5
class LangF(F):
    def resolve_expression(
        self, query=None, allow_joins=True, reuse=None, summarize=False, for_save=False
    ):
        rhs = super().resolve_expression(query, allow_joins, reuse, summarize, for_save)
        if isinstance(rhs.field, TranslatedField):
            field_list = self.name.split("__")
            # TODO: should this always lookup lang
            if len(field_list) == 1:
                # Lookup current lang for one field
                field_list.extend([get_current_language()])
            for name in field_list[1:]:
                # Perform key lookups along path
                rhs = KeyTextTransform(name, rhs)
        return rhs


# TODO: should this just inherit from LangF or do we want one without reference lookups
class L(KeyTextTransform):
    """Expression to return the current language"""

    def __init__(self, *args, **kwargs):
        super().__init__(get_current_language(), *args, **kwargs)
